/*
 * Copyright 2005 Jean-Philippe Aumasson <jpnossa@free.fr>
 * Copyright 2005 Florent Kaisser <jpnossa@free.fr>
 * Copyright 2005 Jean-Baptiste Langlois <florent.kaisser@free.fr>
 *
 * This file is part of SIMS : SIMS Is a Machine Simulator
 *
 * SIMS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SIMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SIMS; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For more details see the file COPYING.
 */
/**
 * create machine and launch simulation in a new thread, called by TmLauncher.
 * @author JP
 * @version 20/11/04
 */
package sims.ihm.mdt;

import sims.*;

import sims.ihm.*;

import sims.ihm.graph.*;

import sims.mdt.*;

import java.awt.*;
import java.awt.event.*;

import java.io.*;

import java.util.*;


public class Simulation extends Thread implements InterfaceSimulation,
    GraphicValues {
    private static final int SLEEP_TIME = 100;
    private MachineRead m;
    private int i = 0;

    // "semaphores" for simulation control
    private boolean stopped = true; // true when machine stopped, false to restart
    private boolean reset = false; // true when reset wanted by user
    private boolean finished = false; // true when simulation finished
    private boolean nextstep = false;
    private boolean running = true;
    private boolean turbo = false; // for turbo mode: wont update graphics
    private String fileName;
    private int speed;
    private TmLauncher launcher; // the main machine window of this simulation
    private JGraph2D gr; // the graph of this simulation

    ////////////////////////////////////////////////////////////////////

    /**
    * Initialize the Simulation.
    * @param fileName the String of the file name.
    * @param l the TmLauncher which created the simulation.
    */
    public Simulation(String fileName, TmLauncher l) {
        this.fileName = fileName;
        launcher = l;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Starts the machine, and keep it in memory after execution to reset it, until viewing is not closed.
    *
    */
    public void run() {
        while (running) // until viewing not closed
         {
            startMachine();

            while ((!reset) && (finished) && (running)) {
                try {
                    sleep(SLEEP_TIME);
                    yield();
                    testReset();
                } catch (InterruptedException e) {
                }
            }
        }
    }

    ////////////////////////////////////////////////////////////////////
    private void startMachine() {
        finished = false;
        launcher.writeMessage("Machine loaded : " + m.getName() +
            ", vitesse: " + speed);

        while (((i == 0) || (i == 3)) && (running)) {
            while ((stopped) && (!nextstep)) // attente "active", pour limiter (bien!) les degats on sleep..


                try {
                    sleep(SLEEP_TIME);
                    yield();
                    testReset(); //active-waiting-like, but mem ok !
                } catch (InterruptedException e) {
                }

            testResetLive();

            nextstep = false; // outside of the loop in case of restart after a stepmode session
            i = m.nextConf();

            if (!turbo) {
                gr.setEtatCourant(m.getCurState());
                gr.repaint();
                launcher.repaint();
                launcher.colorCurrentInstruction(getLine());
            }

            if (i == 1) {
                launcher.writeMessage("Input accepted in " + m.getTimeCplx() +
                    " transitions");
            } else {
                if (i == 2) {
                    if (m.isND()) { // if NT-TM, continue the research
                        i = 0;
                    } else {
                        launcher.writeMessage("Input rejected in " +
                            m.getTimeCplx() + " transitions");
                    }
                } else if (i == 3) {
                    launcher.writeMessage("Breakpoint");
                    stopped = true;
                    launcher.stopControlButton();
                }
            }

            try {
                sleep(VITESSES[speed]);
            } catch (InterruptedException e) {
            }
        }

        finished = true;
        stopped = true;
        launcher.stopControlButton();
        i = 0;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Asks the machine word on the tape i.
    *
    * @param i the number of a tape, from 0 to (number of tapes - 1).
    * @return a String corresponding to the word on the tape i, where the first character is the leftmost non-blank cell visited.
    */
    public String wordOnTape(int i) {
        return m.getTapeWord(i);
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * test if reset is true (button were pushed), and perform actions to reset the machine
    */
    private void testReset() {
        if (reset) {
            try {
                m = new MachineRead(fileName);
            } catch (IOException e) {
            } catch (ExceptionCompilation e) {
                launcher.compilError(e);
            }

            if (gr != null) {
                gr.setMachine(m);
                gr.setEtatCourant(m.getCurState());
                gr.repaint();
            }

            launcher.stopControlButton();
            launcher.repaint();
            reset = false;
            finished = false;
            launcher.writeMessage("Reset !");
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * same as testReset, but don't turn the Go button to red, because this test is done during simulation active
    *
    */
    private void testResetLive() {
        if (reset) {
            try {
                m = new MachineRead(fileName);
            } catch (IOException e) {
            } catch (ExceptionCompilation e) {
                launcher.compilError(e);
            }

            if (gr != null) {
                gr.setMachine(m);
                gr.setEtatCourant(m.getCurState());
                gr.repaint();
            }

            reset = false;
            launcher.writeMessage("Reset !");
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Gives the head position index of the tape i.
    * @param i the number of the tape from 0 to (number of tapes - 1 )
    * @return the scannedSquare position number
    */
    public int headPosition(int i) {
        if (!reset) { // test to avoid error during reset

            return m.getScannedSquare(i);
        }

        return 0;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * @return the current state of the machine.
    *
    **/
    public String currentState() {
        return m.getCurState();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    *
    * @return the number of tapes of the machine.
    */
    public int nbTapes() {
        return m.getNbTapes();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    *
    * Turns the boolean value <tt>stopped</tt> to his contrary, hence stop the machine if running, restart it if stopped.
    */
    public void demarrer() {
        stopped = !stopped;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Same as than demarrer().
    *
    */
    public void arreter() {
        stopped = !stopped;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Reset the machine, by turning the boolean value <tt>reset</tt> at true.
    */
    public void reinitialiser() {
        reset = true;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * To pass to the next step manually.
    *
    */
    public void prochainPas() {
        nextstep = true;
        ;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    *
    * @return the name of the machine.
    */
    public String getNom() {
        return m.getName();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Increases the speed value of one and display a message of the modification, if the speed value was below five.
    * Otherwise display an error message.
    *
    */
    public void incVitesse() {
        if (speed < 5) {
            speed++;
            launcher.writeMessage("speed: " + speed);

            if (speed == 1) {
                turbo = false;
                launcher.writeMessage("turbo mode desactivated");
            }
        } else {
            launcher.writeMessage("max speed reached (5)");
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Decreases the speed value of one, if the speed is at one starts the turbo mode and set speed value to zero.
    * If speed value is zero, display an error message, else display a message for the speed modification..
    *
    */
    public void decVitesse() {
        if (speed > 1) {
            speed--;
            launcher.writeMessage("speed: " + speed);
        } else if (speed == 1) {
            launcher.writeMessage("turbo mode activated");
            turbo = true;
            speed--;
        } else {
            launcher.writeMessage("minimal speed reached (1)");
        }
    }

    public void setVitesse(int i) {
        speed = i;
        launcher.writeMessage("speed: " + i);
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Compile and load a machine from its file name.
    * @param fileName a <tt>String</tt> corresponding to the file name.
    */
    public void charger(String fileName)
        throws IOException, ExceptionCompilation {
        m = null;

        try {
            m = new MachineRead(fileName);
        } catch (IOException e) {
            throw e;
        } catch (ExceptionCompilation e) {
            throw e;
        }

        if (gr != null) {
            gr.setMachine(m);
            gr.setEtatCourant(m.getCurState());
            gr.repaint();
        }

        speed = m.getSpeed();
        finished = false; // to leave the while loop
        launcher.repaint();
        reset = false;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Compile and load a machine from its <tt>File</tt> instance.
    * @param file the File instance corresponding to the file.
    */
    public void charger(File file) throws IOException, ExceptionCompilation {
        m = null;

        try {
            m = new MachineRead("sims/examples/" + file.getName());
        } catch (IOException e) {
        } catch (ExceptionCompilation e) {
            throw e;
        }

        if (gr != null) {
            gr.setMachine(m);
            gr.setEtatCourant(m.getCurState());
            gr.repaint();
        }

        speed = m.getSpeed();
        finished = false; // to leave the while loop
        launcher.repaint();
        reset = false;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Stop the simulation by turning the semaphore <tt>running</tt>
    *
    */
    public void stopp() {
        running = false;
    }

    //////////////////////////////////////////////////////////////////// 

    /**
    *
    * @return <tt>true</tt> if the machine is non deterministic (simple, probabilistic or gravity).
    */
    public boolean isND() {
        return m.isND();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    *
    * @return <tt>true</tt> if the machine is probabilistic (or gravity).
    */
    public boolean isProb() {
        return m.isProb();
    }

    /**
    *
    * @return <tt>true</tt> if the machine is a gravity machine.
    */
    public boolean isGravity() {
        return m.isGravity();
    }

    /**
    * For TmLauncher, to know if machine is running when trying to edit tapes
    * @return the boolean <tt>stopped</tt>, thus <tt>true</tt> if the machine is stopped (or paused), <tt>false</tt> otherwise.
    */
    public boolean isStopped() {
        return stopped;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * cCreate the graph window, from (GraphInterface) m
    *
    */
    public void createGraph() {
        gr = new JGraph2D(m);

        Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
        gr.setLocation(dim.width - GRAPH_WIDTH, 5);

        gr.setVisible(true);
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Close the graph, by setting it non visible and put <tt>gr</tt> at <tt>null</tt>
    */
    public void closeGraph() {
        gr.setVisible(false);
        gr = null;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Show or hide the graph.
    * @param boo if <tt>true</tt> sets the graph visible, otherwise set it non visible.
    */
    public void showGraph(boolean boo) {
        gr.setVisible(boo);
    }

    ////////////////////////////////////////////////////////////////////
    ////// GETTERS

    /**
    * @return the minimum square reached index.
    */
    public int min(int i) {
        if (!reset) {
            return m.getMin(i);
        }

        return 0;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    *
    * @return the current time complexity value, equals to the number of transitions of the machine.
    */
    public int getTime() {
        return m.getTimeCplx();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    *
    * @return the current space complexitu value, equals to the number of cells visited on all the tapes.
    */
    public int getSpace() {
        return m.getSpaceCplx();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    *
    * @return the speed value, from zero (turbo mode) to five.
    */
    public int getVitesse() {
        return speed;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    *
    * @return the number of tapes.
    */
    public int getTapeNb() {
        return m.getNbTapes();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * @return the line number of last used instruction line
    */
    public int getLine() {
        return m.getInstructionLine();
    }

    ////////////////////////////////////////////////////////////////////
    ////// SETTERS

    /**
    * update the content of the tape <tt>i</tt> with <tt>word</tt>, display an error message if update fails.
    *
    * @param tapeNb the number of the tape.
    * @param word the new content of the tape, each caracter for a cell of the tape.
    */
    public void setTapeContent(String word, int tapeNb) {
        if (m.setTapeCont(word, tapeNb) != -1) {
            launcher.repaint();
            launcher.writeMessage("Tape " + tapeNb + " modified: " + word);
        } else {
            launcher.writeMessage("Tape modification failed");
        }
    }
}
