/*
 * Copyright 2005 Jean-Philippe Aumasson <jpnossa@free.fr>
 * Copyright 2005 Florent Kaisser <jpnossa@free.fr>
 * Copyright 2005 Jean-Baptiste Langlois <florent.kaisser@free.fr>
 *
 * This file is part of SIMS : SIMS Is a Machine Simulator
 *
 * SIMS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SIMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SIMS; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For more details see the file COPYING.
 */
/**
*  Main window when a machine loaded: display the editor, message windows, dockable buttons, and load the Simulation.
*
*  @author JP <jpnossa@free.fr>
*  @author florent.kaisser@free.fr
*  @version 20/12/04
*/
package sims.ihm.mdt;

import sims.*;

import sims.ihm.*;

import sims.ihm.graph.*;

import sims.ihm.themes.*;

import sims.mdt.*;

import java.awt.*;
import java.awt.event.*;

import java.io.*;

import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.undo.*;


public class TmLauncher extends JFrame implements WindowListener,
    ActionListener, DocumentListener, GraphicValues, BandW {
    static final private String ENREGISTRER = "save";
    static final private String ENREGISTRER_SOUS = "save as";
    static final private String COMPILER = "compile";
    static final private String EDIT = "edit";
    static final private String OPEN = "open";
    static final private String NEW = "new";
    static final private String QUITTER = "quit";
    static final private String GRAPH_SHOW = "show/hide the graph";
    static final private String TAPES_SHOW = "show/hide the tapes";
    private static final String machinePath = "sims/examples";
    private Simulation sim;
    private TuringView view; // the view window, displaying the tapes content
    private Commandes commands;
    Editeur e;
    private FenetreMessage fm;
    private JToolBar tb;
    private File file;
    private String filePath;
    private JPanel pNord;
    private JPanel pSud;
    private int oldLigne = -1;
    private boolean ok = true;
    private boolean viewLoaded = false;
    private boolean init = false; // turns to true when creating new file/machine
    protected UndoAction undoAction;
    protected RedoAction redoAction;
    protected UndoManager undo = new UndoManager();
    private HashMap actions;
    private JButton bsave;
    private JButton bcompile;
    private JButton bedit;
    private JButton bquit;
    private JToggleButton bgraph;
    private JToggleButton btapes;

    ////////////////////////////////////////////////////////////////////

    /**
    * Initialize the window by creating its graphic components, and load the machine if we chose to open an existing file.
    */
    public TmLauncher(String filePath) {
        this.filePath = filePath;
        file = new File(filePath);

        if (!file.exists()) {
            try {
                file.createNewFile();
                writeDirectives(filePath);
            } catch (IOException ioe) {
                System.out.println(ioe.getMessage());
            }

            init = true;
        }

        sim = new Simulation(filePath, this);
        commands = new Commandes(sim);
        e = new Editeur();
        fm = new FenetreMessage(100, 100, MESS_WIDTH, MESS_HEIGHT);

        try {
            e.read(file);
        } catch (IOException ioe) {
            System.out.println(ioe.getMessage());
        }

        e.setMargin(new Insets(0, 10, 0, 0));
        tb = new JToolBar("Edition commands");
        addButtons();

        // menubar
        creerActionTable(e);

        JMenuBar mb = new JMenuBar();
        mb.add(createFileMenu());
        mb.add(createEditMenu());
        setJMenuBar(mb);
        commands.enableButtons(true);
        pNord = new JPanel(new FlowLayout(FlowLayout.LEFT));
        pSud = new JPanel(new GridLayout(1, 2));

        // add the components
        getContentPane().setLayout(new BorderLayout());
        getContentPane().add(new JScrollPane(e), BorderLayout.CENTER);
        pNord.add(tb);
        pNord.add(commands);
        pSud.add(fm);
        getContentPane().add(pNord, BorderLayout.NORTH);
        getContentPane().add(pSud, BorderLayout.SOUTH);
        e.getDocument().addUndoableEditListener(new MyUndoableEditListener());
        addWindowListener(this);
        pack();
        setBounds(EDIT_X, EDIT_Y, EDIT_WIDTH, EDIT_HEIGHT);
        setVisible(true);

        if (!init) {
            try {
                sim.charger(filePath);

                if (!(sim.getNom() == null)) {
                    setTitle("Edition: " + sim.getNom());
                } else {
                    setTitle("Edition: " + filePath);
                }
            } catch (ExceptionCompilation ec) {
                compilError(ec);
                ok = false;
            } catch (IOException ioe) {
                System.out.println(ioe.getMessage());
                ok = false;
            }

            if (ok) {
                view = new TuringView(sim, this);
                viewLoaded = true;
            }

            sim.createGraph();
            commands.actualiserVitesse();
            sim.start();
        } else {
            setTitle("Edition: name your machine !");
            bgraph.setEnabled(false);
            btapes.setEnabled(false);
            bedit.setEnabled(false);
            commands.enableButtons(false);
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Display a message for syntax errors.
    * @param ec the exception that catched the syntax error.
    *
    */
    public void compilError(ExceptionCompilation ec) {
        fm.addMessage("Syntax error : " + ec.getMessage() + "  la ligne " +
            ec.getLigne());
        e.setLigneSurbrillance(ec.getLigne(), Color.RED);
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Disaply a message in the message window.
    * @param mess the message to display.
    */
    public void writeMessage(String mess) {
        fm.addMessage(mess);
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Called after each transition to color the last instruction used.
    * @param i the current line number
    */
    public void colorCurrentInstruction(int i) {
        e.setLigneSurbrillance(i, Color.YELLOW);
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Reset the <tt>stop & go</tt> button at the end of a calcul.
    */
    public void stopControlButton() {
        if (commands != null) {
            commands.resetButtonGo();
        }
    }

    ////////////////////////////////////////////////////////////////////
    private JToggleButton creerToggleBouton(String imageName,
        String actionCommand, String toolTipText, String altText, boolean etat) {
        //Charge l'image
        String imgLocation = imageName;
        Icon icon = new ImageIcon(imgLocation);

        //cre et initialise le bouton
        JToggleButton button = new JToggleButton();
        button.setActionCommand(actionCommand);
        button.setToolTipText(toolTipText);
        button.addActionListener(this);

        if (icon != null) //image trouv
         {
            button.setIcon(icon);
        } //image non trouv
        else {
            button.setText(altText);
            System.err.println("Resource not found: " + imgLocation);
        }

        button.setSelected(etat);

        return button;
    }

    ////////////////////////////////////////////////////////////////////
    private JButton creerBouton(String imagePath, String actionCommand,
        String toolTipText, String altText) {
        //Charge l'image
        Icon icon = new ImageIcon(imagePath);

        //cre et initialise le bouton
        JButton button = new JButton();
        button.setActionCommand(actionCommand);
        button.setToolTipText(toolTipText);
        button.addActionListener(this);

        if (icon != null) //image trouv
         {
            button.setIcon(icon);
        } //image non trouv
        else {
            button.setText(altText);
            System.err.println("Resource not found: " + imagePath);
        }

        if (imagePath.equals(SAVE_BUTTON)) {
            button.setMnemonic(KeyEvent.VK_S);
        } else if (imagePath.equals(COMPILE_BUTTON)) {
            button.setMnemonic(KeyEvent.VK_K);
        } else if (imagePath.equals(TAPES_BUTTON)) {
            button.setMnemonic(KeyEvent.VK_T);
        } else {
            button.setMnemonic(KeyEvent.VK_Q);
        }

        return button;
    }

    ////////////////////////////////////////////////////////////////////
    private void addButtons() {
        bsave = creerBouton(SAVE_BUTTON, ENREGISTRER, "Save the current file",
                "E");
        bcompile = creerBouton(COMPILE_BUTTON, COMPILER, "Compile the machine",
                "C");
        bedit = creerBouton(TAPES_BUTTON, EDIT, "Tapes edition", "B");
        bgraph = creerToggleBouton(GRAPH_BUTTON, GRAPH_SHOW,
                "Show/hide the graph", "G", false);
        btapes = creerToggleBouton(VIEW_BUTTON, TAPES_SHOW,
                "Show/hide the tapes", "T", false);
        bquit = creerBouton(QUIT_BUTTON, QUITTER, "Quit", "Q");
        tb.add(bsave);
        tb.add(bcompile);
        tb.add(bedit);
        tb.add(bgraph);
        tb.add(btapes);
        tb.add(bquit);
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * hide and deference the graphic components when closing the session for this machine..
    *
    */
    public void closeAll() {
        if (!init) {
            view.setVisible(false);
        }

        setVisible(false);

        if (!init) {
            sim.stopp();
            sim.closeGraph();
            view = null;
        }

        commands = null;
        e = null;
        fm = null;
        sim = null;

        try {
            finalize();
        } catch (Throwable t) {
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Update the view window graphics.
    *
    */
    public void repaint() {
        if (view != null) {
            view.repaint();
        }
    }

    ////////////////////////////////////////////////////////////////////
    /// WindowListener interface redefinitions
    public void windowActivated(WindowEvent e) {
    }

    public void windowClosed(WindowEvent e) {
    }

    public void windowDeactivated(WindowEvent e) {
    }

    public void windowDeiconified(WindowEvent e) {
    }

    public void windowIconified(WindowEvent e) {
    }

    public void windowOpened(WindowEvent e) {
    }

    public void windowClosing(WindowEvent e) {
        closeAll();

        try {
            finalize();
        } catch (Throwable t) {
        }
    }

    ////////////////////////////////////////////////////////////////////
    /// ACTIONS
    private void enregistrer() {
        try {
            e.write(file);
        } catch (IOException ioe) {
        }
    }

    ////////////////////////////////////////////////////////////////////
    private void enregistrerSous() {
        JFileChooser fileSelect = new JFileChooser();
        fileSelect.setCurrentDirectory(file.getParentFile());
        fileSelect.setMultiSelectionEnabled(false);
        fileSelect.addChoosableFileFilter(new TurFilter());
        fileSelect.setAcceptAllFileFilterUsed(false);

        int result = fileSelect.showSaveDialog(this);

        if (fileSelect.getSelectedFile() != null) {
            file = fileSelect.getSelectedFile();
            enregistrer();
            setTitle("Editeur: " + file.getPath());
        }
    }

    ////////////////////////////////////////////////////////////////////
    private void compiler() {
        enregistrer();
        ok = true;

        try {
            sim.charger(filePath);
            setTitle("Edition: " + sim.getNom());
        } catch (ExceptionCompilation ec) {
            compilError(ec);
            ok = false;
        } catch (IOException ioe) {
            System.out.println(ioe.getMessage());
            ok = false;
        }

        if ((ok) && (!viewLoaded)) {
            view = new TuringView(sim, this);
            viewLoaded = true;
        }

        if (init) {
            commands.actualiserVitesse();
            e.updateLigne();
            sim.createGraph();
            sim.start();
            init = false;
            bedit.setEnabled(false);
            bgraph.setEnabled(true);
            btapes.setEnabled(true);
        }

        if (ok) {
            writeMessage("Machine compiled with no error");
            e.enleverSurbrillance();
        }

        commands.enableButtons(ok);
    }

    ////////////////////////////////////////////////////////////////////
    private void editTapes() {
        if (sim.isND()) {
            JOptionPane.showMessageDialog(this,
                "Can't modify tapes of a non-deterministic machine",
                "Sims error", JOptionPane.WARNING_MESSAGE);
        } else if (sim.isStopped()) {
            new TapeUpdate(sim);
        } else {
            JOptionPane.showMessageDialog(this,
                "Can't modify tapes during simulation", "Sims error",
                JOptionPane.WARNING_MESSAGE);
        }
    }

    ////////////////////////////////////////////////////////////////////
    private void openMachine() {
        JFileChooser fileSelect = new JFileChooser();
        fileSelect.setCurrentDirectory(new File(machinePath));
        fileSelect.setMultiSelectionEnabled(false);
        fileSelect.addChoosableFileFilter(new TurFilter());
        fileSelect.setAcceptAllFileFilterUsed(false);

        int result = fileSelect.showOpenDialog(this);

        if ((fileSelect.getSelectedFile() != null) &&
                (result == JFileChooser.APPROVE_OPTION)) { // if we selected a file

            String fileName = fileSelect.getSelectedFile().getPath();

            if (TurFilter.getExtension(fileSelect.getSelectedFile()).equals("tur")) {
                new TmLauncher(fileName);
            }

            closeAll();
        }
    }

    ////////////////////////////////////////////////////////////////////
    private void newMachine() {
        JFileChooser fileSelect = new JFileChooser();
        fileSelect.setCurrentDirectory(new File(machinePath));
        fileSelect.setMultiSelectionEnabled(false);
        fileSelect.addChoosableFileFilter(new TurFilter());
        fileSelect.setAcceptAllFileFilterUsed(false);

        int result = fileSelect.showSaveDialog(this);

        if ((fileSelect.getSelectedFile() != null) &&
                (result == JFileChooser.APPROVE_OPTION)) { // if we selected a file

            String fileName = fileSelect.getSelectedFile().getPath();

            if (!(TurFilter.getExtension(fileSelect.getSelectedFile()).equals("tur"))) {
                // no .tur extension, rename it 

                /*
                 File oldFile = new File(fileName);
                 try {
                 oldFile.createNewFile();
                 } catch (IOException ex) {};
                 */
                fileName = fileName + ".tur";

                /*
                File newFile = new File(fileName);
                boolean b = oldFile.renameTo(newFile);
                */
            }

            //writeDirectives(fileName);
            new TmLauncher(fileName);

            closeAll();
        }
    }

    ////////////////////////////////////////////////////////////////////
    private void writeDirectives(String fname) {
        PrintWriter out = null;
        BufferedWriter buf = null;

        try {
            buf = new BufferedWriter(new FileWriter(fname));
            out = new PrintWriter(buf);
        } catch (IOException ex) {
            System.err.println("err");
        }

        ;

        if (out != null) {
            out.println("/* Enter a description of the machine here */\n");
            out.println("#nom = \"<machine name>\"");
            out.println("#alphabet = \"<alphabet>\"");
            out.println("#nbBandes = <number of tapes>");
            out.println("#accepte = \"<accepting state>\"");
            out.println("#init = \"<initial state>\"");
            out.println("#vitesse = <from 1 to 5>");
            out.println("#bande = \"<initial content of the tape>\"");
            out.println(
                "// #nd = yes // uncomment this line for non-deterministic machine");
            out.println(
                "// #prob = yes // uncomment this line for probabilistic machine");
            out.println(
                "// #gravity = yes // uncomment this line for gravity machine");

            out.close();
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * Redefinition of the <tt>actionListener</tt> method to define instructions to perform when an action occurs.
    *
    */
    public void actionPerformed(ActionEvent ae) {
        String cmd = ae.getActionCommand();

        if (cmd.equals(ENREGISTRER)) {
            enregistrer();
        } else if (cmd.equals(ENREGISTRER_SOUS)) {
            enregistrerSous();
        } else if (cmd.equals(COMPILER)) {
            compiler();
        } else if (cmd.equals(EDIT)) {
            editTapes();
        } else if (cmd.equals(QUITTER)) {
            closeAll();
        } else if (cmd.equals(GRAPH_SHOW)) {
            sim.showGraph(!((JToggleButton) ae.getSource()).isSelected());
        } else if (cmd.equals(TAPES_SHOW)) {
            view.setVisible(!((JToggleButton) ae.getSource()).isSelected());
        } else if (cmd.equals(OPEN)) {
            openMachine();
        } else if (cmd.equals(NEW)) // new machine
         {
            newMachine();
        }
    }

    ////////////////////////////////////////////////////////////////////
    /// INTERFACE DOCUMENTLISTENER
    public void insertUpdate(DocumentEvent e) {
        commands.enableButtons(false);
    }

    public void removeUpdate(DocumentEvent e) {
        commands.enableButtons(false);
    }

    public void changedUpdate(DocumentEvent e) {
        commands.enableButtons(false);
    }

    ////////////////////////////////////////////////////////////////////
    private void creerActionTable(JTextComponent textComponent) {
        actions = new HashMap();

        Action[] actionsArray = textComponent.getActions();

        for (int i = 0; i < actionsArray.length; i++) {
            Action a = actionsArray[i];
            actions.put(a.getValue(Action.NAME), a);
        }
    }

    private Action getActionByName(String name) {
        return (Action) (actions.get(name));
    }

    ////////////////////////////////////////////////////////////////////
    //Create the edit menu.
    protected JMenu createEditMenu() {
        JMenu menu = new JMenu("Edition");
        menu.setMnemonic('E');

        JMenuItem it;

        //Undo and redo are actions of our own creation.
        undoAction = new UndoAction();
        menu.add(undoAction).setAccelerator(KeyStroke.getKeyStroke(
                KeyEvent.VK_Z, InputEvent.CTRL_MASK));
        redoAction = new RedoAction();
        menu.add(redoAction).setAccelerator(KeyStroke.getKeyStroke(
                KeyEvent.VK_Y, InputEvent.CTRL_MASK));
        menu.addSeparator();

        //These actions come from the default editor kit.
        //Get the ones we want and stick them in the menu.
        it = menu.add(getActionByName(DefaultEditorKit.cutAction));
        it.setText("Cut");
        it.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X,
                InputEvent.CTRL_MASK));
        it = menu.add(getActionByName(DefaultEditorKit.copyAction));
        it.setText("Copy");
        it.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C,
                InputEvent.CTRL_MASK));
        it = menu.add(getActionByName(DefaultEditorKit.pasteAction));
        it.setText("Past");
        it.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_V,
                InputEvent.CTRL_MASK));
        menu.addSeparator();
        it = menu.add(getActionByName(DefaultEditorKit.selectAllAction));
        it.setText("Select all");
        it.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_A,
                InputEvent.CTRL_MASK));

        return menu;
    }

    ////////////////////////////////////////////////////////////////////
    //Create the edit menu.
    protected JMenu createFileMenu() {
        JMenu menu = new JMenu("File");
        menu.setMnemonic('F');
        addMenuItem(menu, "New...", NEW);
        addMenuItem(menu, "Open...", OPEN);
        addMenuItem(menu, "Saver", ENREGISTRER);
        addMenuItem(menu, "Save as...", ENREGISTRER_SOUS);
        menu.addSeparator();
        addMenuItem(menu, "Compile", COMPILER);
        menu.addSeparator();
        addMenuItem(menu, "Quit", QUITTER);

        return menu;
    }

    ////////////////////////////////////////////////////////////////////
    private void addMenuItem(JMenu menu, String label, String cmd) {
        /*
                if (label.equals("Nouvelle..."))
                {

                    JMenu newMach = new JMenu(label);
                    JMenuItem dtm = newMach.add("Machine dterministe");
                    JMenuItem ndtm = newMach.add("Machine non-dterministe");
                    JMenuItem ptm = newMach.add("Machine probabiliste");
                    JMenuItem gtm = newMach.add("Machine  gravit");
                    dtm.setActionCommand(DTM);
                    ndtm.setActionCommand(NDTM);
                    ptm.setActionCommand(PTM);
                    gtm.setActionCommand(GTM);
                    dtm.addActionListener(this);
                    ndtm.addActionListener(this);
                    ptm.addActionListener(this);
                    gtm.addActionListener(this);
                    newMach.add(dtm);
                    newMach.add(ndtm);
                    newMach.add(ptm);
                    newMach.add(gtm);
                    menu.add(newMach);
                }
                else
                {
        */
        JMenuItem i = menu.add(label);
        i.setActionCommand(cmd);
        i.addActionListener(this);

        if (label.equals("Save")) {
            i.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S,
                    InputEvent.CTRL_MASK));
        } else if (label.equals("Quit")) {
            i.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q,
                    InputEvent.CTRL_MASK));
        } else if (label.equals("Compile")) {
            i.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_F5, 0));
        } else if (label.equals("Open...")) {
            i.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O,
                    InputEvent.CTRL_MASK));
        } else if (label.equals("New...")) {
            i.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_N,
                    InputEvent.CTRL_MASK));
        }
    }

    ////////////////////////////////////////////////////////////////////
    /// MENUBAR METHODS
    //This one listens for edits that can be undone.
    protected class MyUndoableEditListener implements UndoableEditListener {
        public void undoableEditHappened(UndoableEditEvent e) {
            //Remember the edit and update the menus.
            undo.addEdit(e.getEdit());
            undoAction.updateUndoState();
            redoAction.updateRedoState();
        }
    }

    class UndoAction extends AbstractAction {
        public UndoAction() {
            super("Cancel");
            setEnabled(false);
        }

        public void actionPerformed(ActionEvent e) {
            try {
                undo.undo();
            } catch (CannotUndoException ex) {
                System.out.println("Repeat impossible: " + ex);
                ex.printStackTrace();
            }

            updateUndoState();
            redoAction.updateRedoState();
        }

        protected void updateUndoState() {
            if (undo.canUndo()) {
                setEnabled(true);
                putValue(Action.NAME, undo.getUndoPresentationName());
            } else {
                setEnabled(false);
                putValue(Action.NAME, "Cancel");
            }
        }
    }

    class RedoAction extends AbstractAction {
        public RedoAction() {
            super("Repeat");
            setEnabled(false);
        }

        public void actionPerformed(ActionEvent e) {
            try {
                undo.redo();
            } catch (CannotRedoException ex) {
                System.out.println("Repeat impossible: " + ex);
                ex.printStackTrace();
            }

            updateRedoState();
            undoAction.updateUndoState();
        }

        protected void updateRedoState() {
            if (undo.canRedo()) {
                setEnabled(true);
                putValue(Action.NAME, undo.getRedoPresentationName());
            } else {
                setEnabled(false);
                putValue(Action.NAME, "Repeat");
            }
        }
    }
}
