/*
 * Copyright 2005 Jean-Philippe Aumasson <jpnossa@free.fr>
 * Copyright 2005 Florent Kaisser <jpnossa@free.fr>
 * Copyright 2005 Jean-Baptiste Langlois <florent.kaisser@free.fr>
 *
 * This file is part of SIMS : SIMS Is a Machine Simulator
 *
 * SIMS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SIMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SIMS; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For more details see the file COPYING.
 */
/**
* a TM's tape
*
*  @author JP <jpnossa@free.fr>
*  @version 23/11/04
*/
package sims.mdt;

import sims.ihm.*;

import java.util.*;


public class Tape implements Cloneable, Errors {
    private Alphabet alpha; //tape's alphabet, may be different than machine's one
    private HashMap ruban; // indice (Integer) -> symbole inscrit at this place on the tape (Character)
    private int tapeNb; //  number of the tape, 0 for the main
    private int scannedSquare; //  number of active letter
    private int minimumReached = 0; // negative value if overflow at the left...else 0 (so as to print correctly all the tape)

    ////////////////////////////////////////////////////////////////////

    /**
     *  constructor for the first tape (number zero).
     * @param h the hash table containing the tapes alphabets
     * @param a the alphabet of the machine
     * @param word the entry word
     * */
    public Tape(String word, HashMap h, Alphabet a) {
        this(0, h, a);

        for (int i = 0; i < word.length(); i++) {
            ruban.put(new Integer(i), new Character(word.charAt(i)));
        }
    }

    /**
    *  constructor for any tape except the first
    * @param h the hash table containing the tapes alphabets
    * @param a the alphabet of the machine
    * @param number this tape's number.
    * */
    public Tape(int number, HashMap h, Alphabet a) {
        tapeNb = number;
        ruban = new HashMap();
        alpha = new Alphabet();
        minimumReached = 0;
        scannedSquare = 0;

        // if there's a specific alphabet for this tape
        if (h.containsKey((new Integer(number)).toString())) {
            copyAlphabet((String) h.get((new Integer(number)).toString()));
        } else {
            alpha = a;
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * from the string of the tape number, copy its alphabet from HM TmParser.alphabets
    *
    * */
    private void copyAlphabet(String s) {
        alpha.addLetter(0, ' ');

        for (int i = 0; i < s.length(); i++)
            alpha.addLetter(i + 1, s.charAt(i));
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * @return <tt>true</tt> if the parameter letter belongs to this tape's alphabet.
    * @param c the character we want to test.
    */
    public boolean inAlphabet(Character c) {
        if ((alpha.size() > 0) && (!alpha.contains(c))) {
            return false;
        }

        return true;
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * update the tape: write letter and shift
     *
     * @param s the key
     * */
    protected void update(String s) throws TuringException {
        writeLetter(s);
        doShift(s);
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * update the date: write letter and shift, for a ND Machine
     *
     * @param s the key
     * */
    protected void ndUpdate(String choice) {
        writeLetter(choice);
        doShift(choice);
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * writes the letter corresponding to the first char of the String s
     *
     * @param s a value of hashtable moves: letter.shift
     */
    private void writeLetter(String s) {
        if (inAlphabet(new Character(s.charAt(0)))) {
            ruban.put(new Integer(scannedSquare), new Character(s.charAt(0)));
        } else {
            TuringException.error("The symbol " + s.charAt(0) +
                " on the tape " + tapeNb + " does not belong to its alphabet");
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * read the moving letter and shift consequently the head
     *
     * @param s a value of hashtable moves: < letter.shift >
     */
    private void doShift(String s) {
        String t = s.substring(2);

        if (t.compareToIgnoreCase("L") == 0) {
            leftShift();
        } else if (t.compareToIgnoreCase("R") == 0) {
            rightShift();
        } else if (t.compareToIgnoreCase("N") == 0) {
        } else {
            TuringException.error(error10 + t);
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
     *  right-shift the tape head
     */
    private void rightShift() {
        scannedSquare++;

        if (!ruban.containsKey(new Integer(scannedSquare))) {
            ruban.put(new Integer(scannedSquare), new Character(' '));
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
     *  left-shift the tape head
     */
    private void leftShift() {
        scannedSquare--;

        if ((scannedSquare < 0) && (scannedSquare < minimumReached)) {
            minimumReached = scannedSquare;
        }

        if (!ruban.containsKey(new Integer(scannedSquare))) {
            ruban.put(new Integer(scannedSquare), new Character(' '));
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * redefinition needed to clone a <tt>Tape</tt>
    *
    */
    public Object clone() throws CloneNotSupportedException {
        Object o = super.clone();

        ((Tape) o).setTapeStatus(ruban);

        return o;
    }

    ////////////////////////////////////////////////////////////////////
    ////// TOOL METHODS FOR ND-D SIMULATION

    /**
    * Set the tape content as the parameter.
    * @param rub a hash table describing a tape, taking as key an index value (signed <tt>Integer</tt>)
    */
    public void setTapeStatus(HashMap rub) {
        ruban = (HashMap) rub.clone();
    }

    ////////////////////////////////////////////////////////////////////
    ////// GETTERS
    ////////////////////////////////////////////////////////////////////

    /**
     *  @return the letter fronting the head
     *
     */
    public Character getScannedSymbol() {
        if (ruban.containsKey(new Integer(scannedSquare))) // if indexed
         {
            return (Character) ruban.get(new Integer(scannedSquare));
        }

        return new Character(' ');
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * @return the scanned square index.
    */
    public int getScannedSquare() {
        return scannedSquare;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * @return the minimum cell reached (maximum value is zero, because it's the starting index).
    */
    public int getMin() {
        return minimumReached;
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * @return the word on the tape, from head to last non-blank char
     */
    public String getWord() {
        int i = scannedSquare;
        StringBuffer s = new StringBuffer();

        while (ruban.containsKey(new Integer(i))) {
            if (ruban.get(new Integer(i)).equals(new Character(' '))) {
                break;
            }

            s.append((Character) ruban.get(new Integer(i)));
            i++;
        }

        return s.toString();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    *  @return the tape content from leftmost to rightmost square visited (blanks included)
    *
    */
    public String getTapeContent() {
        StringBuffer s = new StringBuffer();

        if (ruban.size() > 0) {
            for (int i = minimumReached; i < (ruban.size() + minimumReached);
                    i++) {
                s.append((Character) ruban.get(new Integer(i)));
            }
        } else {
            s.append(' ');
        }

        return s.toString();
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * @return the content of the tape, from minimumReached to last cell accessed
     */
    public String getStatus() {
        int i = minimumReached; // pour afficher la sortie, on doit avoir i = scannedSquare
        StringBuffer s = new StringBuffer();

        while (ruban.containsKey(new Integer(i))) {
            // => Va automatiquemt jusqu'au maxmodifi
            s.append((Character) ruban.get(new Integer(i)));
            i++;
        }

        return s.toString();
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * @return the number of cells visited, equals number of cells indexed in the hash <tt>ruban</tt>.
     *
     */
    public int getSpace() {
        return ruban.size();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * @return the character indexed at <tt>i</tt> if exists, <tt>null</tt> otherwise.
    */
    public Character getAlphaLetter(int i) {
        return alpha.getLetter(i);
    }

    ////////////////////////////////////////////////////////////////////

    /**
    *
    * @return the tape's alphabet, instance of <tt>Alphabet</tt>
    */
    public Alphabet getAlpha() {
        return alpha;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    *
    * @return the alphabet size, equals to the number of letters.
    */
    public int getAlphaSize() {
        return alpha.size();
    }

    ////////////////////////////////////////////////////////////////////
    ////// SETTERS

    /**
    * Add a letter to this tape's alphabet.
    * @param i the index where to place the letter.
    * @param c the letter to add.
    */
    public void setAlpha(int i, char c) {
        alpha.addLetter(i, c);
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * set the tape content, from its current min, as the specified String
    *
    */
    public void setTapeContent(String s) {
        int j = minimumReached;
        ruban.clear();

        for (int i = 0; i < s.length(); i++) {
            ruban.put(new Integer(j), new Character(s.charAt(i)));
            j++;
        }
    }
}
