/*
 * Copyright 2005 Jean-Philippe Aumasson <jpnossa@free.fr>
 * Copyright 2005 Florent Kaisser <jpnossa@free.fr>
 * Copyright 2005 Jean-Baptiste Langlois <florent.kaisser@free.fr>
 *
 * This file is part of SIMS : SIMS Is a Machine Simulator
 *
 * SIMS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SIMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SIMS; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For more details see the file COPYING.
 */
/**
 * class used to read a "*.tur" file, descripting a turing machine, and inserts informations in memory structures
 *
*  @author JP <jpnossa@free.fr>
*  @version 20/11/04
 */
package sims.mdt;

import sims.*;

import sims.ihm.*;

import java.io.*;

import java.text.StringCharacterIterator;

import java.util.*;


public class TmParser implements Errors {
    private static final int MAX_LINES = 50; // max number of empty lines between 2 declarations
    private String fileName;
    private HashMap alphabets; // cl i => tape i's alphabet (String), if exists
    private Vector instructions; // vector of string containing each line of instruction
    private Vector ordInst; // vector containing vectors of each instruction component (String)
    private MachineRead m;
    private int linesRead = -1; // number of lines read in the file, used to localise errors.
    private HashMap instructionLine; // for each instruction, return its line

    ////////////////////////////////////////////////////////////////////

    /**
    * Initialize the parser.
    * @param inFile a string of the file to parse name.
    * @param m the <tt>MachineRead</tt> instance associated to the parser.
    *
    */
    public TmParser(String inFile, MachineRead m)
        throws IOException, ExceptionCompilation {
        fileName = inFile;
        this.m = m;
        instructions = new Vector();
        ordInst = new Vector();
        instructionLine = new HashMap();
        alphabets = new HashMap();
        parse();
        parseInst();
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * read file content according to the syntax
     *
     * */
    private void parse() throws IOException, ExceptionCompilation {
        BufferedReader lecteur = null;
        int i = 0;
        int p = 1;
        String s = new String();

        try {
            lecteur = new BufferedReader(new FileReader(fileName));
        } catch (FileNotFoundException ex) {
            new TuringException(error1 + fileName);
        }

        while (i < MAX_LINES) {
            s = lecteur.readLine().trim();

            linesRead++;

            if (s.equals("")) {
            } else {
                switch (s.charAt(0)) {
                case '#':

                    try {
                        copyDirective(s);
                    } catch (TuringException f) {
                    }

                    ;

                    break;

                case '/':

                    switch (s.charAt(1)) {
                    case '/':
                        break; // line-comment, nothing, next-line

                    case '*':

                        while (!((s.charAt(p - 1) == '*') &&
                                (s.charAt(p) == '/'))) {
                            if (p == (s.length() - 1)) { // if end of line
                                s = lecteur.readLine().trim();
                                linesRead++;
                                p = 0;
                            }

                            while (s.equals("")) { // skip empty lines
                                s = lecteur.readLine().trim();
                                linesRead++;
                            }

                            p++;
                        }

                        p = 0;

                        break;
                    }

                    ;

                    break;

                case '(':
                    i = MAX_LINES;

                    break; // instruction found..leave the loop            

                default:
                    break;
                }
            }

            i++;
        }

        // insertion des instructions dans Vector instructions
        int b = 0; // counter

        // si on a detect une parenthse auparavant
        if (s.length() > 1) { // on test dabord si on a une ligne nulle

            if (s.trim().charAt(0) == '(') {
                instructions.add(s.trim());
                instructionLine.put(new Integer(b), new Integer(linesRead));
                b++;
            }
        }

        while (((s = lecteur.readLine()) != null)) { // for each instruction line
            linesRead++;

            if (!(s.trim().equals(""))) { // if non-empty line

                if (s.trim().charAt(0) == '(') { // considering line as an instruction if first char is '('
                    instructions.add(s.trim());
                    instructionLine.put(new Integer(b), new Integer(linesRead));
                    b++;
                } else if (s.trim().charAt(0) == '/') {
                    switch (s.charAt(1)) {
                    case '/':
                        break; // line-comment, nothing, next-line

                    case '*':

                        while (!((s.charAt(p) == '*') &&
                                (s.charAt(p + 1) == '/'))) {
                            if (p == (s.length() - 1)) { // if end of line
                                s = lecteur.readLine().trim();
                                p = 0;
                            }

                            while (s.equals("")) {
                                s = lecteur.readLine().trim();
                            }

                            p++;
                        }

                        p = 0;

                        break;

                    default:
                        throw new ExceptionCompilation(linesRead, error2);
                    }
                } else {
                    throw new ExceptionCompilation(linesRead, error2);
                }
            }
        }

        // close the stream
        lecteur.close();

        // we won't need them anymore, so leave it to the GC
        lecteur = null;
        fileName = null;
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * copy directive name and its value at the good memory struct (hash table)
     *
     * @param s a directive line
     */

    // TODO: manage all directives
    private void copyDirective(String s)
        throws ExceptionCompilation, TuringException {
        String z = nameDirective(s);
        String w;
        w = directiveValue(s);

        if ((w.equals("")) && (z.equals("bande"))) {
            m.setInWord(w);
        } else if (w.equals("")) // if no value to the directive -> Grer exception
         {
            throw new TuringException(error3 + z);
        }

        if (z.equals("init")) {
            m.setInitState(withoutQuotes(w));
        } else if (z.equals("nom")) {
            m.setName(withoutQuotes(w));
        } else if (z.equals("nbBandes")) {
            try {
                m.setNbTapes(Integer.parseInt(w));
            } catch (NumberFormatException nb) {
                throw new ExceptionCompilation(linesRead, error14);
            }

            ;
        } else if (z.equals("bande")) {
            m.setInWord(withoutQuotes(w));
        } else if (z.equals("alphabet")) {
            if (isTapeAlphabet(w)) {
            } else {
                try {
                    for (int i = 0; i < withoutQuotes(w).length(); i++)
                        m.setAlpha(i + 1, withoutQuotes(w).charAt(i));
                } catch (ExceptionCompilation e) {
                    throw e;
                }
            }
        } else
        // on ajoute chaque char de la chaine a la hashtable
        // REFAIRE POUR GERER LA VIRG COMME SEPARATEUR
        // ACCEP UN SEUL ETAT (plusieurs chars possible)
        if (z.equals("accepte")) {
            for (int i = 0; i < w.split(",").length; i++)
                m.setStateStatus(withoutQuotes(w.split(",")[i]), "acc");
        } else if (z.equals("rejette")) {
            for (int i = 0; i < w.split(",").length; i++)
                m.setStateStatus(withoutQuotes(w.split(",")[i]), "rej");
        } else if (z.equals("stop")) {
            for (int i = 0; i < w.split(",").length; i++)
                m.setStateStatus(withoutQuotes(w.split(",")[i]), "stop");
        } else if (z.equals("nd")) {
            if (isYes(w)) {
                m.setChoiceMachine();
            }
        } else if (z.equals("prob")) {
            if (isYes(w)) {
                m.setProbMachine();
                m.setChoiceMachine();
            }
        } else if (z.equals("gravity")) {
            m.setProbMachine();
            m.setChoiceMachine();
            m.setGravity();
        } else if (z.equals("vitesse")) {
            try {
                if ((Integer.parseInt(w) > 0) && (Integer.parseInt(w) < 6)) {
                    m.setSpeed(Integer.parseInt(w));
                } else {
                    throw new ExceptionCompilation(linesRead, error4);
                }
            } catch (NumberFormatException e) {
                throw new ExceptionCompilation(linesRead, error4);
            }
        }
    }

    ////////////////////////////////////////////////////////////////////
    private boolean isYes(String w) {
        if ((w.compareToIgnoreCase("o") == 0) ||
                (w.compareToIgnoreCase("y") == 0) ||
                (w.compareToIgnoreCase("yes") == 0) ||
                (w.compareToIgnoreCase("oui") == 0)) {
            return true;
        }

        return false;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * from a directive value, return true if it's an alphabet tape (<tapenb>,<alphabet>)
    *
    * */
    private boolean isTapeAlphabet(String s) throws ExceptionCompilation {
        if (s.indexOf(',') > 0) { // if thers a comma, then it's a tape alphbt

            String[] parts = s.trim().split(",");

            // creation of alphabet of the tape concerned
            try {
                alphabets.put(parts[0], withoutQuotes(parts[1]));
            } catch (ExceptionCompilation e) {
                throw e;
            }

            ;

            return true;
        }

        // ELSE traitement error
        return false;
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * returns the value without it quotes, if has some
     * */
    private String withoutQuotes(String s) throws ExceptionCompilation {
        if (s.charAt(0) == '"') {
            if (s.charAt(s.length() - 1) == '"') {
                return s.substring(1, s.length() - 1);
            } else {
                throw new ExceptionCompilation(linesRead, error5);
            }
        } else if (s.charAt(s.length() - 1) != '"') {
            return s;
        } else {
            throw new ExceptionCompilation(linesRead, error6);
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * delete comments on the line s (supposed already trim()-ed )
     * @param s a line read from *.tur file
     * @return line without comments
     * */
    private String noComment(String s) {
        for (int i = 0; i < (s.length() - 1); i++)
            if ((s.charAt(i) == '/') && (s.charAt(i + 1) == '/')) {
                return s.substring(0, i).trim();
            }

        return s;
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * return the directive name corresponding to the line s
     *
     * @param s a directive line
     */
    private String nameDirective(String s) {
        if (!(s.charAt(0) == '#')) {
            return "";
        }

        for (int i = 0; i < s.trim().length(); i++)
            if (s.trim().charAt(i) == '=') {
                return s.substring(1, i - 1);
            }

        return "";
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * return the value corresponding to the directive line s
     *
     * @param s a directive line
     */
    private String directiveValue(String s) {
        for (int i = 0; i < s.trim().length(); i++)
            if (s.trim().charAt(i) == '=') {
                return noComment((s.trim().substring(i + 1).trim()));
            }

        return "";
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * from vector instructions containing raw instructions,
     * create Vector of Vector containing for each inst a vector of its components ( card = 4*nbTapes + 2 )
     *
     *
     * */
    private void parseInst() throws ExceptionCompilation {
        for (int i = 0; i < instructions.size(); i++) // for each instructions
         {
            ordInst.add(new Vector()); // create the new Vector 

            insertValues(i); // insert correctly the values included in the String in ordInst
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * from instruction string line, parse and copy its elements into a vector, element of ordInst
     * @param i number of current instruction line
     * */
    private void insertValues(int i) throws ExceptionCompilation {
        String[] elements;
        int k = ((String) instructions.get(i)).indexOf('(');
        int j = ((String) instructions.get(i)).indexOf(')');

        if (k == -1) {
            throw new ExceptionCompilation(((Integer) instructionLine.get(
                    new Integer(i))).intValue(), error11);
        }

        if (j == -1) {
            throw new ExceptionCompilation(((Integer) instructionLine.get(
                    new Integer(i))).intValue(), error12);
        }

        String s = ((String) instructions.get(i)).substring(k + 1, j); // s <- contentof parenthesis w/o em

        elements = s.split(",");

        try {
            if (checkInstruction(elements, i)) {
                for (int l = 0; l < elements.length; l++) {
                    ((Vector) ordInst.get(i)).add(elements[l]);
                }
            }
        } catch (TuringException e) {
        }

        ;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * check if an instruction is correct (shift, letters, size)
    * @param el an instruction to check
    * */
    private boolean checkInstruction(String[] el, int i)
        throws TuringException {
        int ok = 0;
        String mess = "Incorrect instruction line " +
            (Integer) instructionLine.get(new Integer(i));

        if (m.getNbTapes() == 1) {
            if (el.length != 5) {
                throw new TuringException(mess + error7);
            } else if (((el[3].compareToIgnoreCase("L") == 0) &&
                    (el[3].compareToIgnoreCase("N") == 0) &&
                    (el[3].compareToIgnoreCase("R") == 0)) ||
                    (!m.inAlphabet(el[1].charAt(0))) ||
                    (!m.inAlphabet(el[2].charAt(0)))) {
                throw new TuringException(mess + error8);
            } else {
                return true;
            }
        } else { // more than 1 tape

            for (int j = 1; j <= m.getNbTapes(); j++)
                if (el.length == ((4 * j) + 2)) {
                    ok = 1;
                }

            if (ok != 1) {
                throw new TuringException(mess);
            } else { // { well-sized tape)

                for (int j = 0; j < ((el.length - 2) / 4); j++) // (length-2)/4 = nber of tapes in this instruction


                    if (((el[(4 * j) + 4].compareToIgnoreCase("L") == 0) &&
                            (el[(4 * j) + 4].compareToIgnoreCase("N") == 0) &&
                            (el[(4 * j) + 4].compareToIgnoreCase("R") == 0)) ||
                            (!m.inAlphabet(el[(4 * j) + 2].charAt(0)) ||
                            (!m.inAlphabet(el[(4 * j) + 3].charAt(0))))) {
                        throw new TuringException(mess + error8);
                    }
            }
        }

        return true;
    }

    ////////////////////////////////////////////////////////////////////
    ////// GETTERS

    /**
     * @return the instruction at line <tt>i</tt>
     */
    public String getInstructions(int i) {
        return ((String) instructions.get(i));
    }

    /**
    * @return the number of instructions read.
    */
    public int getNbInstructions() {
        return instructions.size();
    }

    /**
    * @return the (String) start state/vertex for the i-th instruction
    */
    public String getStartVert(int i) {
        return ((String) instructions.get(i)).split(",")[0].substring(1);
    }

    /**
    * @return the ending state (end vertex in the graph) of the instruction <tt>i</tt>.
    * @param i an index of an instruction.
    */
    public String getEndVert(int i) {
        String inst = (String) instructions.get(i);
        inst = inst.substring(0, inst.indexOf(')') + 1); // to avoid comments problems...juste select from '(' to ')'

        String[] s = (inst).split(",");
        String t = s[s.length - 1]; //get the last part of the split result (~ new state)

        return t.substring(0, t.indexOf(')'));
    }

    /**
    * @return the graph-label of the i-th transition
    */
    public String getLabel(int i) {
        // create s as the instruction line without its parenthesis
        String s = ((String) instructions.get(i));
        s = s.substring(0, s.indexOf(')'));
        s = s.substring(1);

        String[] t = s.split("[,]");
        StringBuffer buf = new StringBuffer();

        for (int j = 0; j < t.length; j++) {
            buf.append(t[j]);
        }

        return s;
    }

    /**
    * @return the number of states
    */
    public int getNbStates() {
        // for each instruction get the start and end states, if unknown add em to the list
        Vector vertices = new Vector();

        for (int i = 0; i < getNbInstructions(); i++) {
            if (!vertices.contains(getStartVert(i))) {
                vertices.add(getStartVert(i));
            }

            if (!vertices.contains(getEndVert(i))) {
                vertices.add(getEndVert(i));
            }
        }

        return vertices.size();
    }

    /**
     *
     * @return the jth component of instruction i
     *
     * @param i indice of instruction line
     * @param j indice of component in the line
     * */
    public String getInstComp(int i, int j) {
        return (String) ((Vector) ordInst.get(i)).get(j);
    }

    /**
     * @return the size (nb of elements) of instruction i
     * */
    public int getInstCard(int i) {
        return ((Vector) ordInst.get(i)).size();
    }

    /**
    * @return the <tt>HashMap</tt> of the alphabets read.
    */
    public HashMap getAlphabet() {
        return alphabets;
    }

    /**
    * @return the number of the line from its index.
    */
    public int getLine(int nbinst) {
        return ((Integer) instructionLine.get(new Integer(nbinst))).intValue();
    }
}
