/*
 * Copyright 2005 Jean-Philippe Aumasson <jpnossa@free.fr>
 * Copyright 2005 Florent Kaisser <jpnossa@free.fr>
 * Copyright 2005 Jean-Baptiste Langlois <florent.kaisser@free.fr>
 *
 * This file is part of SIMS : SIMS Is a Machine Simulator
 *
 * SIMS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SIMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SIMS; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For more details see the file COPYING.
 */
/**
* Turing's Machine main class
*
*  @author JP <jpnossa@free.fr>
*  @version 20/12/04
*/
package sims.mdt;

import sims.*;

import sims.ihm.*;

import sims.ihm.graph.*;

import sims.ihm.mdt.*;

import sims.mdt.nd.*;

import java.io.*;

import java.util.*;


public abstract class TuringMachine implements Errors, GrapheInterface {
    protected Alphabet alpha; // Alphabet of the machine, maybe diff than each tape's alphabet
    protected NDMachine ndMachine; // the associated NDTM if the machine is ND.
    protected Vector rubans; // Vector containing the tapes (instances of Tape)
    protected Vector unsizedKeys; // Vector to save the (String) short keys
    protected Vector ndQueue; // indeterminisation vector, contains the undetermination choices
    protected Vector selectionner; // vector containing ndQueue's indexes..used for G-Machine
    protected HashMap moveState; // (String) <etat.tapenb.letter.tapenb2.letter....> -> (String) <etat>
    protected HashMap moveStateUnsized; // special map containing transition with special size ( < 4*nbTapes + 2 )
    protected HashMap stateStatus; // (String) <etat> ->  acc | rej | stop
    protected HashMap ndMoveState; // HashMap of Vectors, to contain each following states, in addition of the one in tranState (unsized included)
    protected HashMap weights; // used by a G-machine; (String)state -> (Integer)its weight.
    protected MConfig currentConfBackup; // to backup a m-configuration
    protected TmParser p; // parser associated, needed here to get instructions' line numbers
    protected String name; // this tm's name
    protected String curState = "q_0"; // current state, default initial state = q_0
    protected String inWord; // the entry word of the machine
    protected String initState;
    protected int nbTrans = 0; // number of transitions realized
    protected int nbSpace = 0; // number of cells visited
    protected int nbTapes; // nber of tapes (indexed in (Vector)rubans from 0 to (nbTapes-1)
    protected int speed = 1; // from 1 to 5
    protected int currentLine; // line nber in the file of last instruction used
    protected boolean choiceMachine = false; // true if non-deterministic machine
    protected boolean probMachine = false; // true if probabilist machine..=> choiceMachine = true
    protected boolean gMachine = false; // true if gravity-machine (see /docs for details on G-machines)

    ////////////////////////////////////////////////////////////////////

    /**
     *  Only constructor, initialize structures.
     * */
    public TuringMachine() throws IOException {
        alpha = new Alphabet();
        alpha.addLetter(0, ' ');
        rubans = new Vector();
        moveState = new HashMap();
        stateStatus = new HashMap();
        moveStateUnsized = new HashMap();
        unsizedKeys = new Vector();
    }

    ////////////////////////////////////////////////////////////////////

    /**
     *  calculs next config, in both cases of an deterministic of non deterministic machine by calling the appropriate method.
     *
     *  @return 0 if nothing special, 1 for acception, else 2 if refused
     * */
    public int nextConf() {
        if (gMachine) {
            addWeight(); // increment the weight of the current state
        }

        // done here and not after, cos' only useful when used by a transition, values does not matter at end of calcul
        if (!choiceMachine) {
            return nextConfD();
        }

        return nextConfND();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * calculs next configuration, for deterministic machine
    *  @return 0 if nothing special, 1 for acception, else 2 if refused
    */
    public int nextConfD() {
        String s; // ~ string corresponding to the current config: State.symbolsScanned...

        if (fullSizedInstr()) // test if no multisized instruction fits
         {
            s = buildConfigString();

            try {
                updateState(s); // ~ curState <- new state
                updateTapes(s); // update each tape (shift + write symbol)
            } catch (TuringException e) {
            }
        }

        return returnStatus();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * nextConf for a nondeterministic machine (executed even if no indetermination presently faced)
    *  @return 0 if nothing special, 1 for acception, else 2 if refused
    */
    public int nextConfND() {
        int j = -1; // initial value of polysized matchingkey
        int detDegree = 0; // degr d'indetermination, ++ pour chaque transition possible
        int transChose = 0; // for the prob simulation, randome-generated number, <= detDegree
        String s = " "; // the m-config string

        if (ndQueue.size() == 0) // on n'a pas de transition ND  explorer
         {
            // load the config at fifo's head
            loadHeadConfig();

            // build the config-string
            s = buildConfigString();
            detDegree = detDegree(s);

            if (detDegree > 1) // if more than 1 trans possible, -> ND
             {
                enqueueTransitions(s); // for the states and tapes, enqueue those transitions in their ndQueue
                backupCurrentConfig();
            } else if (detDegree == 1) {
                if (fullSizedInstr()) // si une instruction variable correspond  la situationt
                 {
                    try {
                        updateState(s); // curState <- new state
                        updateTapes(s); // update each tape (shift + letter writing)
                    } catch (TuringException e) {
                    }
                }

                // put the created config into the Fifo
                enqueueCurrentConf();

                return returnStatus();
            } else if (detDegree == 0) {
                System.out.println("config crash: " + s);
                TuringException.error(error9);
            }
        } // if there, remain some transitions to explore for a given config
        else {
            restoreConfig(); // restore the config concerned
        }

        // if probabilist machine, generate the randomized choice
        if (probMachine) {
            if (gMachine) {
                transChose = ((Integer) selectionner.get(Math.round(
                            (selectionner.size() - 1) * (float) Math.random()))).intValue();
            } else {
                transChose = Math.round((detDegree - 1) * (float) Math.random());
            }
        }

        // creation of a new config: update state and tapes
        ndUpdateState(((Instruction) ndQueue.get(transChose)).getNewState()); // if non-prob, transChose=0, head of queue
        ndUpdateTapes((Instruction) ndQueue.get(transChose));
        currentLine = (((Instruction) ndQueue.get(transChose)).getLine());

        if (probMachine) // if probabilistic, clear the others choices, don't care of them...
         {
            ndQueue.clear();
        } // remove the transition simulated, others automatically left-shifted
        else {
            ndQueue.remove(0);
        }

        // add the current MConfig to the Fifo, so as to explore it later
        enqueueCurrentConf();

        return returnStatus();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * finally called by nextConf-like methods ,
    * to return a value corresponding to the (new) status of the machine, between:
    * accepted, refused, stop, nothing special
    *
    */
    private int returnStatus() {
        if (stateStatus.containsKey(curState)) // test new current state status
         {
            if (stateStatus.get(curState).equals("acc")) {
                return 1;
            } else if (stateStatus.get(curState).equals("rej")) {
                return 2;
            } else if (stateStatus.get(curState).equals("stop")) {
                return 3;
            } else {
                return 0;
            }
        }

        return 0;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * build the String describing the current Config:
    * state.tapenumber0.symbolread0.tapenumber1.(...).symbolreadN
    */
    private String buildConfigString() {
        StringBuffer s = new StringBuffer();
        s.append(curState);
        s.append(".");

        for (int k = 0; k < (nbTapes - 1); k++) {
            s.append(Integer.toString(k));
            s.append(".");
            s.append(((Tape) rubans.get(k)).getScannedSymbol().toString());
            s.append(".");
        }

        s.append(Integer.toString(nbTapes - 1));
        s.append(".");
        s.append(((Tape) rubans.get(nbTapes - 1)).getScannedSymbol().toString());

        return s.toString();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * called by nextConfND: calculs the degree of an indetermination (nber of possible transt)
    */
    private int detDegree(String s) {
        int i = 0;

        if (moveState.containsKey(s)) {
            i++;
        }

        i += existsUnFullSized().size(); // == 0 if no instruction matching the curconf

        if (ndMoveState.containsKey(s)) // if there are, add to detDeg the number of matching transtns
         {
            i += ((Vector) ndMoveState.get(s)).size();
        }

        return i;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * called by nextConfND: enqueue the transitions into the ndQueue:
    *  new State in machine's ndQueue, key for tapes' "moves" HM, in theirs ndQueue.
    *  priority order: moveState, ndMoveState, non-fullsized
    */
    private void enqueueTransitions(String s) {
        Vector j;
        Instruction buf;

        if (gMachine) {
            selectionner.clear();
        }

        /*
        Sauver dans un Vector index(= index dans ndQueue) -> weight de l'tat

        Vector selection
         valeurs -> indices de ndQueue, et si le poids = n, on rajoute au vector
         n entres avec comme value l'indice ndQueu
         et pour trouver la trans choisie on fera un random sur la taille du vector
        */
        if (moveState.containsKey(s)) {
            buf = (Instruction) moveState.get(s);
            ndQueue.add(buf);

            if (gMachine) {
                addSelectionWeight(buf.getNewState(), 0);
            }
        }

        if (ndMoveState.containsKey(s)) {
            for (int i = 0; i < ((Vector) ndMoveState.get(s)).size(); i++) {
                buf = (Instruction) ((Vector) ndMoveState.get(s)).get(i);
                ndQueue.add(buf);

                if (gMachine) {
                    addSelectionWeight(buf.getNewState(), ndQueue.size() - 1);
                } // ndQueue.size() - 1 = this Instr index in ndQueue
            }
        }

        j = existsUnFullSized(); // j = vector of (String) keys (= multisized instr)

        if (j.size() > 0) // if we have some non-full-sized-instr matching
         {
            for (int i = 0; i < j.size(); i++) {
                buf = (Instruction) moveStateUnsized.get((String) (j.get(i)));
                ndQueue.add(buf);

                if (gMachine) {
                    addSelectionWeight(buf.getNewState(), ndQueue.size() - 1);
                }
            }
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * insert into selectionner N entries where N = nwState's weight
    * associated values i are the ndQueue index of the transition.
    *
    */
    private void addSelectionWeight(String nwState, int i) {
        if (weights.containsKey(nwState)) { // if has a weight value

            for (int k = 0; k < ((Integer) weights.get(nwState)).intValue();
                    k++) {
                selectionner.add(new Integer(i));
            }
        } else {
            selectionner.add(new Integer(i));
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * update the state of the machine, by searching for a matching config string in moveState
     * @param s String representing the current configuration
     * */
    private void updateState(String s) throws TuringException {
        // if current config matches an entry of the transition table
        if (moveState.containsKey(s)) {
            curState = ((Instruction) moveState.get(s)).getNewState(); // update state
            currentLine = ((Instruction) moveState.get(s)).getLine();
            nbTrans++;
        } else {
            throw new TuringException("key " + s + " unknown");
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * update all the tapes of the machine
     * */
    private void updateTapes(String s) throws TuringException {
        for (int i = 0; i < nbTapes; i++) {
            ((Tape) rubans.get(i)).update(((Instruction) moveState.get(s)).getTapeTrans(
                    i));
        }
    }

    ////////////////////////////////////////////////////////////////////
    private void ndUpdateState(String s) {
        curState = s;
        nbTrans++;
    }

    ////////////////////////////////////////////////////////////////////
    private void ndUpdateTapes(Instruction choice) {
        for (int i = 0; i < nbTapes; i++) {
            if ((choice.getTapeTrans(i)) != null) // test if this trans is defined for this tape
             {
                ((Tape) rubans.get(i)).ndUpdate(choice.getTapeTrans(i));
            }
        }
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * search for an variable instruction matching our config, if there is, update the machine, and return false
     *
     * */
    private boolean fullSizedInstr() {
        String s = new String();
        Vector matchingKey = existsUnFullSized();
        int nbt;
        int k;

        //  if no instruction matching found
        if (matchingKey.size() == 0) {
            return true;
        }

        // { instruction found }, update the machine
        s = (String) matchingKey.get(0);

        int sep = s.indexOf('.');
        curState = ((Instruction) moveStateUnsized.get((String) matchingKey.get(
                    0))).getNewState();
        currentLine = ((Instruction) moveStateUnsized.get((String) matchingKey.get(
                    0))).getLine();

        try {
            for (int j = 0; j < (((s.length() - 1 + sep) - 1) / 4); j++) //(s.length()-1)/4 = nb of tapes concerned
             {
                nbt = (Integer.parseInt(s.substring(((4 * j) + 2 + sep) - 1,
                            ((4 * j) + 3 + sep) - 1)));
                ((Tape) rubans.get(nbt)).update(((Instruction) moveStateUnsized.get(
                        s)).getTapeTrans(nbt));
            }
        } catch (TuringException e) {
        }

        nbTrans++;

        return false;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * looks for keys in memory matching our current config,
    * if there are, return a vector of keys.
    * @return a vector of matching keys
    */
    private Vector existsUnFullSized() {
        String s = new String();
        Vector matchingKeys = new Vector();
        int sep; // first separator position

        // for each instrucion of moveStateUnsized: test state, then the symbols reads on each tape
        for (int i = 0; i < unsizedKeys.size(); i++) {
            s = (String) unsizedKeys.get(i);
            sep = s.indexOf('.');

            if (s.substring(0, sep).equals(curState)) // if state is matching, continue analysis
             {
                for (int j = 0; j < (((s.length() - 1 + sep) - 1) / 4); j++) {
                    // if the tape number char(4*j+1) has as scannedSymbol char at 4*j+2, ok, otherwise break
                    if (!(((Tape) rubans.get(Integer.parseInt(s.substring(((4 * j) +
                                        2 + sep) - 1, ((4 * j) + 3 + sep) - 1)))).getScannedSymbol()
                               .toString().equals(s.substring(((4 * j) + 4 +
                                    sep) - 1, ((4 * j) + 5 + sep) - 1)))) {
                        break;
                    }

                    // if j == ((..., all tests successfully passed, thus ok for that key, we add it
                    if (j == ((((s.length() - 1 + sep) - 1) / 4) - 1)) {
                        matchingKeys.add(unsizedKeys.get(i));
                    }
                }
            }
        }

        return matchingKeys;
    }

    ////////////////////////////////////////////////////////////////////

    /**
     * @return the word written on the main tape (i), from the first non-blank char to the last non-blank
     *
     * */
    public String readWord(int i) {
        return ((Tape) rubans.get(i)).getWord();
    }

    ////////////////////////////////////////////////////////////////////
    ////// TOOLS FOR LOADING / SAVING A M-CONFIGURATION

    /**
    * backup the current config (Tapes + curState) in currentConfBackup
    * the tapes are cloned
    *
    */
    private void backupCurrentConfig() {
        currentConfBackup = null;
        currentConfBackup = currentConf();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    *
    * @return the current config, instance of <tt>MConfig</tt>
    */
    public MConfig currentConf() {
        Tape[] tapes = new Tape[nbTapes]; // structure to save the tape instances

        for (int i = 0; i < nbTapes; i++)
            tapes[i] = (Tape) rubans.get(i);

        return new MConfig(tapes, curState, nbTrans);
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * restore a config stored in currentConfBackup
    *
    */
    private void restoreConfig() {
        rubans.clear();

        for (int i = 0; i < nbTapes; i++) {
            rubans.add(i, currentConfBackup.getTape(i));
        }

        curState = currentConfBackup.getState().toString();
        nbTrans = currentConfBackup.getTime();
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * load the MConfig at Fifo's head, and removes it of the Fifo
    *
    */
    private void loadHeadConfig() {
        MConfig c = ndMachine.loadConfig();
        rubans.clear();
        rubans = null;
        rubans = new Vector();

        for (int i = 0; i < nbTapes; i++) {
            rubans.add(i, c.getTape(i));
        }

        curState = c.getState().toString();
        nbTrans = c.getTime();
        c = null;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * enqueue the current conf into the Fifo
    *
    */
    protected void enqueueCurrentConf() {
        ndMachine.enqueueConfig(currentConf());
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * To know if the machine is non deterministic. Will return <tt>true</tt> if the machine is probabilistic, or gravity, those are special cases of non determinism.
    * @return <tt>true</tt> if the machine is non deterministic, <tt>false</tt> otherwise.
    */
    public boolean isND() {
        return choiceMachine;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * To know if a machine is probabilistic, will return <tt>true</tt> if the machine is a gravity one.
    * @return <tt>true</tt> if the machine is probabilistic, <tt>false</tt> otherwise.
    */
    public boolean isProb() {
        return probMachine;
    }

    ////////////////////////////////////////////////////////////////////

    /**
    * To know if a machine is a gravity machine.
    * @return <tt>true</tt> if the machine is a gravity machine, <tt>false</tt> otherwise.
    */
    public boolean isGravity() {
        return gMachine;
    }

    ////////////////////////////////////////////////////////////////////
    private void addWeight() {
        if (weights.containsKey(curState)) {
            Integer i = (Integer) weights.get(curState);
            weights.put(curState, new Integer(i.intValue() + 1));
        } else {
            weights.put(curState, new Integer(2)); // 2 cos' we consider weight=1 if !containsKey
        }
    }

    ////////////////////////////////////////////////////////////////////
    //// FOR GRAPHEINTERFACE

    /**
    * Will return a vector of edges, one for each instruction in the file.
    * @return a collection of edges, instances of <tt>Arc</tt>.
    */
    public Collection arcs() {
        Vector v = new Vector();

        for (int i = 0; i < p.getNbInstructions(); i++) {
            v.add(new Arc(p.getStartVert(i), p.getEndVert(i), p.getLabel(i)));
        }

        return v;
    }

    ////////////////////////////////////////////////////////////////////
    public int getNbrEtat() {
        return p.getNbStates();
    }

    ////////////////////////////////////////////////////////////////////
    public int getNbrArc() {
        return p.getNbInstructions();
    }

    public Object getRacine() { // racine du graphe

        return initState;
    }

    ////////////////////////////////////////////////////////////////////
    ////// GETTERS

    /**
    * @return the tape content, as a <tt>String</tt>.
    */
    public String getTapeStatus(int i) {
        return ((Tape) rubans.get(i)).getStatus();
    }

    /**
    * @return the number of tapes of the machine.
    */
    public int getNbTapes() {
        return nbTapes;
    }

    /**
    * @return the current state.
    */
    public String getCurState() {
        return curState;
    }

    /**
    * @return an <tt>Integer</tt> corresping to the status of the state passed as parameter. 1 if the state is accepting, 2 if it's rejecting, 3 for a breakpoint, 0 for others.
    */
    public Integer getStateStatus(String key) {
        return (Integer) stateStatus.get(key);
    }

    /**
    * @return the new state corresponding to the key passed in parameter, for full sized deterministic instructions.
    */
    public String getTransState(String key) {
        return (String) moveState.get(key);
    }

    /**
    * @return the current space complexity, equals to the number of cells visited on all tapes.
    */
    public int getSpaceCplx() {
        int c = 0;

        for (int i = 0; i < rubans.size(); i++) {
            c += ((Tape) rubans.get(i)).getSpace();
        }

        return c;
    }

    /**
    * @return the current time complexity value, equals to the number of transitions done.
    */
    public int getTimeCplx() {
        return nbTrans;
    }

    /**
    * @return the speed of this machine simulation.
    */
    public int getSpeed() {
        return speed;
    }

    /**
    * @return the word on the tape <tt>i</tt>.
    */
    public String getTapeWord(int i) {
        return ((Tape) rubans.get(i)).getTapeContent();
    }

    /**
    * @return the scanned square index for the tape <tt>i</tt>.
    */
    public int getScannedSquare(int i) {
        return (((Tape) rubans.get(i)).getScannedSquare()); // +
    }

    /**
    * @return the name of the machine.
    */
    public String getName() {
        return name;
    }

    /**
    * @return the minimal cell reached on the tape <tt>i</tt>.
    */
    public int getMin(int i) {
        return ((Tape) rubans.get(i)).getMin();
    }

    /**
    * @return the line of the last used instruction.
    */
    public int getInstructionLine() {
        return currentLine;
    }

    ////////////////////////////////////////////////////////////////////
    ////// SETTERS

    /**
    * To set the tape <tt>i</tt> content as the <tt>String</tt> in parameter.
    * @param s the word we want to copy on the tape, replacing the old one.
    * @param i the tape we want to update number
    * @return 1 if operation succeeded, -1 if an error occured.
    */
    public int setTapeCont(String s, int i) {
        boolean ok = true;

        for (int k = 0; k < s.length(); k++) {
            if (!(((Tape) rubans.get(i)).inAlphabet(new Character(s.charAt(k))))) {
                new TuringException(error13 + i);
                ok = false;
            }
        }

        if (ok) {
            ((Tape) rubans.get(i)).setTapeContent(s);

            return 1;
        }

        return -1;
    }
}
